from typing import Any, Tuple

from a3json_struct.errors import ValidationError
from a3json_struct import validators
from .abstract_field import AbstractField
from .utils import JsonType, OpenAPIFormat


class FloatField(AbstractField):
    
    def __init__(
            self,
            min_value: float = None,
            max_value: float = None,
            *args, **kwargs
    ):
        super().__init__(*args, **kwargs)

        self._min_value = min_value
        self._max_value = max_value

        if self._min_value is not None:
            self._validators.append(validators.MinValueValidator(self._min_value))
        if self._max_value is not None:
            self._validators.append(validators.MaxValueValidator(self._max_value))

    def _cast_to_python(self, value: Any) -> float:
        if isinstance(value, float):
            return value

        try:
            return float(str(value))
        except (TypeError, ValueError):
            raise ValidationError(f'Value "{value}" is not a valid float number.')

    def _cast_to_json(self, cleaned_value: float) -> float:
        return cleaned_value

    def _get_json_type_and_openapi_format(self) -> Tuple[str, str]:
        return JsonType.Number, OpenAPIFormat.Float

    def generate_meta_object(self) -> dict:
        od = super().generate_meta_object()

        od['min_value'] = self._min_value
        od['max_value'] = self._max_value

        return od



# -*- coding: utf-8 -*-
from a3json_struct.json_struct import JsonStruct

from a3json_struct.fields.boolean_field import BooleanField
from a3json_struct.fields.char_field import CharField
from a3json_struct.fields.date_field import DateField
from a3json_struct.fields.datetime_field import DateTimeField
from a3json_struct.fields.decimal_field import DecimalField
from a3json_struct.fields.float_field import FloatField
from a3json_struct.fields.integer_field import IntegerField
from a3json_struct.fields.list_field import ListField
from a3json_struct.fields.object_field import ObjectField

from a3json_struct.fields.int_date_field import IntDateField
from a3json_struct.fields.dict_field import DictField



# -*- coding: utf-8 -*-
from .max_length_validator import MaxLengthValidator
from .max_value_validator import MaxValueValidator
from .min_length_validator import MinLengthValidator
from .min_value_validator import MinValueValidator
from .regex_validator import RegexValidator



from typing import Any, Tuple
import datetime

from a3json_struct.errors import ValidationError
from .abstract_field import AbstractField
from .utils import JsonType, OpenAPIFormat


class DateTimeField(AbstractField):

    def _cast_to_python(self, value: Any) -> datetime.datetime:
        if isinstance(value, datetime.datetime):
            return value

        try:
            return datetime.datetime.fromisoformat(str(value))
        except ValueError:
            raise ValidationError(
                f'Value "{value}" must has the correct format (YYYY-MM-DD HH:MM[:ss[.uuuuuu]][TZ])，"'
                f'and must be a valid datetime.'
            )

    def _cast_to_json(self, cleaned_value: datetime.datetime) -> str:
        return cleaned_value.isoformat()

    def _get_json_type_and_openapi_format(self) -> Tuple[str, str]:
        return JsonType.String, OpenAPIFormat.Date



from typing import Any, Set, Tuple
from a3json_struct.errors import ValidationError
from a3json_struct import validators

from .abstract_field import AbstractField
from .utils import JsonType, OpenAPIFormat, set_nonempty_kv


class IntegerField(AbstractField):

    def __init__(
            self,
            min_value: int = None,
            max_value: int = None,
            choices: Set[int] = None,
            *args, **kwargs
    ):
        super().__init__(*args, **kwargs)

        self._min_value = min_value
        self._max_value = max_value
        self._choices = choices

        if self._min_value is not None:
            self._validators.append(validators.MinValueValidator(self._min_value))
        if self._max_value is not None:
            self._validators.append(validators.MaxValueValidator(self._max_value))

    def _validate(self, value: Any) -> Any:
        value = super()._validate(value)
        if self._choices is not None and value is not None and value not in self._choices:
            raise ValidationError(f'Value "{value}" is not a valid choice.')
        return value

    def _cast_to_python(self, value: Any) -> int:
        if isinstance(value, int):
            return value

        try:
            return int(value)
        except (TypeError, ValueError):
            raise ValidationError(f'Value "{value}" is not a valid integer number.')

    def _cast_to_json(self, cleaned_value: int) -> int:
        return cleaned_value

    def _get_json_type_and_openapi_format(self) -> Tuple[str, str]:
        return JsonType.Number, OpenAPIFormat.Long

    def generate_openapi_object(self) -> dict:
        od = super().generate_openapi_object()

        set_nonempty_kv(od, 'minimum', self._min_value)
        set_nonempty_kv(od, 'maximum', self._max_value)

        if self._choices is not None:
            od['enum'] = list(self._choices)

        return od

    def generate_meta_object(self) -> dict:
        od = super().generate_meta_object()

        od['min_value'] = self._min_value
        od['max_value'] = self._max_value
        od['choices'] = list(self._choices) if self._choices is not None else None

        return od





class ValidationError(Exception):

    def __init__(self, message: str):
        self._message = message
        self._field_name_list = list()
        self._index_list = list()

    def set_index(self, index: int):
        self._index_list.insert(0, index)

    def set_field(self, field_name: str):
        if len(self._index_list) > 0:
            for index in self._index_list:
                field_name += f'[{index}]'
            self._index_list = list()

        self._field_name_list.insert(0, field_name)

    def __str__(self) -> str:
        if len(self._field_name_list) > 0:
            field_name = '.'.join(self._field_name_list)
            error_message = f"{field_name}: {self._message}"
        else:
            error_message = self._message

        return error_message



from typing import Any, Tuple

from a3json_struct.errors import ValidationError
from .abstract_field import AbstractField
from .utils import JsonType, OpenAPIFormat


class BooleanField(AbstractField):

    def _cast_to_python(self, value: Any) -> bool:
        if isinstance(value, bool):
            return value

        value = str(value).lower()

        if value in ('t', 'true', '1'):
            return True

        if value in ('f', 'false', '0'):
            return False

        raise ValidationError(f'Value "{value}" is not a valid boolean.')

    def _cast_to_json(self, cleaned_value: bool) -> bool:
        return cleaned_value

    def _get_json_type_and_openapi_format(self) -> Tuple[str, str]:
        return JsonType.Boolean, OpenAPIFormat.Boolean



# -*- coding: utf-8 -*-
from typing import Any


def set_nonempty_kv(d: dict, key: str, v: Any):
    if v not in [None, '']:
        d[key] = v


class JsonType:
    Boolean = 'boolean'
    String = 'string'
    Number = 'number'
    Array = 'array'
    Object = 'object'


class OpenAPIFormat:
    Boolean = 'boolean'
    String = 'string'
    Date = 'date'
    Datetime = 'datetime'
    Decimal = 'decimal'
    Object = 'object'
    Float = 'float'
    Long = 'long'
    Array = 'array'



from typing import Dict, Any, Type
import copy

from a3py.simplified.dynamic import find_all_subclasses
from a3py.simplified.case import snake2camel
from a3json_struct.errors import ValidationError
from a3json_struct.fields.abstract_field import AbstractField
from a3json_struct.fields.utils import JsonType


class JsonStructMetaClass(type):

    def __new__(mcs, name, bases, attrs: dict, **kwargs):
        # exclude self
        parents = [b for b in bases if isinstance(b, JsonStructMetaClass)]
        if not parents:
            return super().__new__(mcs, name, bases, attrs, **kwargs)

        # `field_name_set`: subclass's field first, if parent class has the same field, it will be ignored
        field_name_set = set()
        fields = dict()
        for field_name, field_instance in attrs.items():
            field_name_set.add(field_name)

            if isinstance(field_instance, AbstractField):
                fields[field_name] = field_instance

        # pop the fields of current class
        for field_name in fields.keys():
            attrs.pop(field_name)

        # new class
        new_cls = super().__new__(mcs, name, bases, attrs, **kwargs)

        # parent fields
        mro_structs = new_cls.mro()
        if len(mro_structs) > 3:
            parent_structs = list()
            for kls in mro_structs[1: -2]:
                if issubclass(kls, JsonStruct) and kls != JsonStruct:
                    parent_structs.append(kls)

            for parent in parent_structs:
                parent_fields = parent.get_fields()

                for name, field in parent_fields.items():
                    new_field = copy.deepcopy(field)
                    if name not in field_name_set:
                        fields[name] = new_field
                        field_name_set.add(name)

        # bind to class
        for field_name, field_instance in fields.items():
            field_instance.contribute_to_struct(new_cls, field_name)

        # set private _fields
        new_cls._fields = fields
        return new_cls


class JsonStruct(metaclass=JsonStructMetaClass):
    _fields: Dict[str, AbstractField]
    _has_full_clean: bool

    @classmethod
    def get_fields(cls) -> Dict[str, AbstractField]:
        return cls._fields

    def __init__(self, **kwargs):
        self._set_has_full_clean(False)

        field_name_list = list()
        for field_name, field_instance in self.get_fields().items():
            field_name_list.append(field_name)
            self._super_setattr(field_name, None)

        for k, v in kwargs.items():
            if k in field_name_list:
                setattr(self, k, v)

    def _super_setattr(self, name: str, value: Any):
        super().__setattr__(name, value)

    def _set_has_full_clean(self, v: bool):
        self._super_setattr('_has_full_clean', v)

    def __setattr__(self, name: str, value: Any):
        self._super_setattr(name, value)

        if self._has_full_clean:
            self._set_has_full_clean(False)

    def __str__(self) -> str:
        name = self.__class__.__name__
        if not self._has_full_clean:
            name = f"{name}['not-clean']"
        return name

    def _clean_fields(self):
        for field_name, field_instance in self.get_fields().items():
            value = getattr(self, field_name, None)

            try:
                value = field_instance.clean(value)
            except ValidationError as e:
                e.set_field(field_name)
                raise e

            value = field_instance.clean(value)
            setattr(self, field_name, value)

    def _clean_struct(self):
        # for custom override
        pass

    def full_clean(self):
        self._clean_fields()
        self._clean_struct()
        self._set_has_full_clean(True)

    def to_json(self) -> dict:
        if not self._has_full_clean:
            self.full_clean()

        rd = dict()
        for field_name, field_instance in self.get_fields().items():
            value = getattr(self, field_name, None)
            rd[field_name] = field_instance.to_json(value)

        return rd

    @classmethod
    def generate_openapi_schema(cls) -> dict:
        required_list = list()
        properties = dict()

        for field_name, field_instance in cls.get_fields().items():
            if field_instance.is_required():
                required_list.append(field_name)

            properties[field_name] = field_instance.generate_openapi_object()

        return {
            'type': JsonType.Object,
            'properties': properties,
            'required': required_list,
        }

    @classmethod
    def generate_meta_schema(cls) -> dict:
        fields = dict()

        for field_name, field_instance in cls.get_fields().items():
            fields[field_name] = field_instance.generate_meta_object()

        return {
            'fields': fields
        }

    @classmethod
    def build_variant_from_meta_schema(cls, schema: Dict[str, Dict], class_name: str = None) -> Type['JsonStruct']:
        fields = dict()
        for field_name, meta_object in schema['fields'].items():
            instance = _get_field_instance_by_meta(cls, meta_object)
            fields[field_name] = instance

        class_name = class_name or 'Variant'
        # noinspection PyTypeChecker
        variant: Type['JsonStruct'] = type(class_name, (cls, ), dict())
        setattr(variant, '_fields', fields)
        return variant


_all_inner_fields_classes: Dict[str, Type] = dict()


def _get_all_inner_field_class() -> Dict:
    if len(_all_inner_fields_classes) == 0:
        field_class_list = find_all_subclasses('a3json_struct.struct', AbstractField)
        for field_class in field_class_list:
            _all_inner_fields_classes[field_class.__name__] = field_class

    return _all_inner_fields_classes


def _get_field_instance_by_meta(base: Type['JsonStruct'], meta_object: dict) -> AbstractField:
    inner_classes = _get_all_inner_field_class()

    class_name = meta_object.pop('class_name')
    field_class = inner_classes.get(class_name)
    assert field_class is not None, f"Unknown inner field class: {class_name}"

    if class_name == 'ListField':
        element_field_meta = meta_object.pop('element_field_meta')
        meta_object['element_field'] = _get_field_instance_by_meta(base, element_field_meta)
    elif class_name == 'ObjectField':
        obj_kls_meta = meta_object.pop('obj_kls_meta')
        obj_kls_name = meta_object.pop('obj_kls_name')
        meta_object['obj_kls'] = base.build_variant_from_meta_schema(obj_kls_meta, class_name=snake2camel(obj_kls_name))

    instance = field_class(**meta_object)
    return instance



from typing import Any, Tuple
import datetime

from a3json_struct.errors import ValidationError
from .abstract_field import AbstractField
from .utils import JsonType, OpenAPIFormat


class DateField(AbstractField):

    def _cast_to_python(self, value: Any) -> datetime.date:
        if isinstance(value, datetime.date):
            return value

        try:
            return datetime.date.fromisoformat(str(value))
        except ValueError:
            raise ValidationError(f'Value "{value}" must has the correct format (YYYY-MM-DD)，and must be a valid date.')

    def _cast_to_json(self, cleaned_value: datetime.date) -> str:
        return cleaned_value.isoformat()

    def _get_json_type_and_openapi_format(self) -> Tuple[str, str]:
        return JsonType.String, OpenAPIFormat.Date



from typing import Any, Set, Tuple
from a3json_struct.errors import ValidationError
from a3json_struct import validators

from .abstract_field import AbstractField
from .utils import JsonType, OpenAPIFormat, set_nonempty_kv


class CharField(AbstractField):

    def __init__(
            self,
            min_length: int = None,
            max_length: int = None,
            choices: Set[str] = None,
            pattern: str = None,
            *args, **kwargs
    ):
        super().__init__(*args, **kwargs)

        self._min_length = min_length
        self._max_length = max_length
        self._choices = choices
        self._pattern = pattern

        if self._min_length is not None:
            self._validators.append(validators.MinLengthValidator(self._min_length))
        if self._max_length is not None:
            self._validators.append(validators.MaxLengthValidator(self._max_length))
        if self._pattern is not None:
            self._validators.append(validators.RegexValidator(self._pattern))

    def _validate(self, value: Any) -> Any:
        value = super()._validate(value)
        if self._choices is not None and value is not None and value not in self._choices:
            raise ValidationError(f'Value "{value}" is not a valid choice.')
        return value

    def _cast_to_python(self, value: Any) -> str:
        if isinstance(value, str):
            return value
        else:
            return str(value)

    def _cast_to_json(self, cleaned_value: str) -> str:
        return cleaned_value

    def _get_json_type_and_openapi_format(self) -> Tuple[str, str]:
        return JsonType.String, OpenAPIFormat.String

    def generate_openapi_object(self) -> dict:
        od = super().generate_openapi_object()

        set_nonempty_kv(od, 'minLength', self._min_length)
        set_nonempty_kv(od, 'maxLength', self._max_length)
        set_nonempty_kv(od, 'pattern', self._pattern)

        if self._choices is not None:
            od['enum'] = list(self._choices)

        return od

    def generate_meta_object(self) -> dict:
        od = super().generate_meta_object()

        od['min_length'] = self._min_length
        od['max_length'] = self._max_length
        od['choices'] = list(self._choices) if self._choices is not None else None
        od['pattern'] = self._pattern

        return od



import abc
from typing import Any, Callable, List, Type, Tuple

from a3json_struct.errors import ValidationError
from .utils import set_nonempty_kv


class AbstractField(abc.ABC):

    def __init__(
            self,
            verbose_name: str = None,
            default: Any = None,
            required: bool = True,
            validators: List[Callable] = None,
            description: str = None
    ):
        self._verbose_name = verbose_name
        self._default = default
        self._required = required
        self._validators = validators or list()
        self._description = description
        # struct
        self._struct_kls = None
        self._name = None

    def __str__(self) -> str:
        if self._struct_kls is None:
            return super().__str__()
        else:
            return f"{self._struct_kls.__name__}.{self._name}"

    def contribute_to_struct(self, struct_kls: Type, name: str):
        self._struct_kls = struct_kls
        self._name = name
        if self._verbose_name is None:
            self._verbose_name = name

    def _validate(self, value: Any) -> Any:
        if value is None and self._default is not None:
            if isinstance(self._default, Callable):
                value = self._default()
            else:
                value = self._default

        if value is None and self._required:
            raise ValidationError("This field is required.")

        return value

    def _run_validators(self, value: Any):
        for v in self._validators:
            v(value)

    def clean(self, value: Any) -> Any:
        value = self._validate(value)
        if value is None:
            return

        value = self._cast_to_python(value)

        self._run_validators(value)
        return value

    @abc.abstractmethod
    def _cast_to_python(self, value: Any) -> Any:
        raise NotImplementedError()

    @abc.abstractmethod
    def _cast_to_json(self, cleaned_value: Any) -> Any:
        raise NotImplementedError()

    def to_json(self, cleaned_value: Any) -> Any:
        if cleaned_value is None:
            return
        else:
            return self._cast_to_json(cleaned_value)

    def is_required(self) -> bool:
        return self._required

    @abc.abstractmethod
    def _get_json_type_and_openapi_format(self) -> Tuple[str, str]:
        raise NotImplementedError()

    def generate_openapi_object(self) -> dict:
        json_type, openapi_format = self._get_json_type_and_openapi_format()
        od = {
            'type': json_type,
            'format': openapi_format,
        }

        set_nonempty_kv(od, 'title', self._verbose_name)
        set_nonempty_kv(od, 'description', self._description)

        if self._default is not None:
            if isinstance(self._default, Callable):
                value = self._default()
            else:
                value = self._default

            od['default'] = self.to_json(value)

        return od

    def generate_meta_object(self) -> dict:
        # It is used only for simple verification, so it's not support custom validators or default.
        return {
            'class_name': self.__class__.__name__,
            'verbose_name': self._verbose_name,
            'required': self._required,
            'description': self._description
        }
